"""
BioAt BedTools
# todo
"""

from bioat.lib.libdataclasses import Bed


class BedTools:
    """Bed toolbox."""

    def __init__(self):
        pass

    def random(
        self,
        kind="human",
        length=200,
        number=100,
        fasta=None,
        seed=0,
        rm_n=True,
    ) -> Bed:
        """Get random region

        :param kind: Species:Tag, defaults to "human"
        :type kind: str, optional
        :param length: random region length, defaults to 200
        :type length: int, optional
        :param number: random region number, defaults to 100
        :type number: int, optional
        :param fasta: get random region from which genome, must be defined when rm_n=True, defaults to None
        :type fasta: str, optional
        :param seed: random seed, defaults to 0
        :type seed: int, optional
        :param rm_n: remove random region contains N base (It's often the telomere region), defaults to True
        :type rm_n: bool, optional
        :return: bed file
        :rtype: Bed
        """
        # TODO
